/*  multi_serwer.c
 *  
 *  (C) Copyright I 2005 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *
 *
 * Thanks to Jakub 'darkjames' Zawadzki for pointing me that
 * IPV6_V6ONLY :)
 */

#include <stdarg.h>
#define __USE_BSD
#include <stdio.h>
#include <stdlib.h>		/* exit() */
#include <string.h>		/* str*() */
#include <sys/types.h>		/* socket() */
#define __USE_POSIX		/* Linux kill() */
#define __EXTENSIONS__		/* Solaris kill() */
#include <signal.h>		/* signal()                        */
#include <sys/socket.h>		/* bind(), listen(), accept()      */
#include <sys/wait.h>		/* waitpid()                       */
#include <unistd.h>		/* fork()                          */
#include <ctype.h>		/* is*() */

#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sched.h>

#if !defined(FIONBIO) && !defined(O_NONBLOCK) && !defined(O_NDELAY)
#include <sys/filio.h>		/* FIONBIO on Solaris */
#endif

#ifdef HAVE_LIBWRAP
#include <tcpd.h>
#endif

int gatoi(char **buf, int *a);

void sigchld_handler (int sig);
void sigint_handler(int sig);
void sigint_child_handler(int sig);

int makeitnonblocking (int fd);
int clearselect (void);

int odbieramy (int x, struct sockaddr *osin, int *osinlen);

int closesockets(void);
int shutdown_server(void);

int dodawanie (int a);
int mnozenie  (int a);
int leetspeech(int a);

int potomek(int serv, int fd);
int gprint(char *msg, ...);

typedef struct g_simlist_ {
	char *string;
	struct g_simlist_ * next;
} *g_simlist;
g_simlist add_to_queue(int i, char *string);


#define MBLOCK 4096
#define MAXCLIENTS 5

/* AS IT IS GLOBAL VARIABLE IT'LL BE INITIALIZED WITH 0'S 
 * K&R2 Sec. 4.9
 */
typedef struct g_sokecik_ {
	union {
		int	port;
		char	path[120];
	} dst;
	int family; /* unix, inet, inet6 */
	int proto; /* raw, stream, dgram */
	int (*handler)(int);
	char daemon[100];

	union {
		struct sockaddr_in	ipv4, clie_4;
		struct sockaddr_in6	ipv6, clie_6;
		struct sockaddr_un	unices, clie_u;
	} sa;
	struct sockaddr	*psa, *pclsa;
	int struct_size, cl_len;
	int sd;
	int klient;
	int dzieci[MAXCLIENTS];
/* TO MAKE IT EASIER, ONE BUFFER FOR EACH CONNECTION
 * TAKE A LOOK ON COMMENT BELOW [TYPE '/(bufory' IN VI(M)]
 */
	char bufor[MBLOCK];
	g_simlist kolejka_out;
	g_simlist kolejka_out_tail;
	char pbuf;
} g_sokecik;

#define S_IPV4 sizeof(struct sockaddr_in)
#define S_IPV6 sizeof(struct sockaddr_in6)
#define S_UNIX sizeof(struct sockaddr_un)

g_sokecik ski[] = {
   { {.port = 10000 },		PF_INET,  SOCK_STREAM, dodawanie,  "add"   },
   { {.port = 10000 },		PF_INET,  SOCK_DGRAM,  mnozenie,   "mul"   },
   { {.port = 20000 },		PF_INET6, SOCK_STREAM, mnozenie,   "mul6"  },
   { {.port = 20000 },		PF_INET6, SOCK_DGRAM,  leetspeech, "eleet6"},

   { {.path = "/tmp/ustr" },	PF_UNIX,  SOCK_STREAM, leetspeech, "eleet" },
   { {.path = "/tmp/udgm" },	PF_UNIX,  SOCK_DGRAM,  mnozenie,   "mul"   }
};
const int PORTS_NO = sizeof(ski)/sizeof(g_sokecik);

fd_set rd, wr;
int highsock;

char nor[]="acegiloqstz";
char lit[]="4<36!109572";

int prepare_and_bind_sockets(void)
{
	int i, sockopt;
	for (i=0; i<PORTS_NO; i++)
	{
		ski[i].sd = socket (ski[i].family, ski[i].proto, 0);
		if (ski[i].sd < 0) {
			perror(" (G!) socket()");
			gprint("-socket [%d] won't be used!\n", i);
			ski[i].sd = -1;
		}
	}

	gprint("+Socket(s) created\n");

	for (i=0; i<PORTS_NO; i++)
	{
		if (ski[i].sd == -1)
			continue;

		if (ski[i].family == PF_INET)
		{
			printf("    - %d inet  %s %d\t%s\n", i, 
					ski[i].proto==SOCK_STREAM?"stream":"dgram ",
					ski[i].dst.port, ski[i].daemon);
			ski[i].sa.ipv4.sin_addr.s_addr	= INADDR_ANY;
			ski[i].sa.ipv4.sin_family	= ski[i].family;
			ski[i].sa.ipv4.sin_port		= htons (ski[i].dst.port);
			ski[i].psa = (struct sockaddr *)&ski[i].sa.ipv4;
			ski[i].pclsa = (struct sockaddr *)&ski[i].sa.clie_4;
			ski[i].cl_len = ski[i].struct_size = S_IPV4;
		} else if (ski[i].family == PF_INET6) {
			printf("    - %d inet6 %s %d\t%s\n", i, 
					ski[i].proto==SOCK_STREAM?"stream":"dgram ",
					ski[i].dst.port,ski[i].daemon);
			ski[i].sa.ipv6.sin6_addr	= in6addr_any;
			ski[i].sa.ipv6.sin6_family	= ski[i].family;
			ski[i].sa.ipv6.sin6_port	= htons (ski[i].dst.port);
			ski[i].psa = (struct sockaddr *)&ski[i].sa.ipv6;
			ski[i].pclsa = (struct sockaddr *)&ski[i].sa.clie_6;
			ski[i].cl_len = ski[i].struct_size = S_IPV6;

			/* THIS IS FOR LINUX COMPATIBILITY
			 * IN SOLARIS NOT NEEDED
			 *
			 * WITHOUT THAT YOU'LL GET AN ERROR WHEN DOING
			 * listen:
			 *  [+] ON IPv6 SOCKET [IF IPv4 SOCKET WAS
			 *      CREATED EARLIER _ON THE SAME PORT_]
			 *  [+] ON IPv4 SOCKET [IF IPv6 SOCKET WAS
			 *      CREATED EARLIER _ON THE SAME PORT_]
			 *
			 * THIS HAPPENS BECAUSE LINUX BY DEFAULT TRIES
			 * TO ENCAPSULATE IPV4 IN IPV6...
			 */
#ifdef IPV6_V6ONLY
			if (setsockopt (ski[i].sd, IPPROTO_IPV6, IPV6_V6ONLY,
					(char *)&sockopt, sizeof(sockopt)) < 0)
			{
				perror (" (G!) setsockopt: IPV6_ONLY");
				gprint("-socket [%d] won't be used!\n", i);
				close(ski[i].sd);
				ski[i].sd = -1;
				continue;
			}
#endif
		} else {
			printf("    - %d unix  %s %s\t%s\n", i,
					ski[i].proto==SOCK_STREAM?"stream":"dgram ",
					ski[i].dst.path, ski[i].daemon);
			ski[i].sa.unices.sun_family	= ski[i].family;
			/* yeah I know, it'd be better to do strncpy
			 */
			unlink(ski[i].dst.path);
			strcpy(ski[i].sa.unices.sun_path, ski[i].dst.path);
			ski[i].psa = (struct sockaddr *)&ski[i].sa.unices;
			ski[i].pclsa = (struct sockaddr *)&ski[i].sa.clie_u;
			ski[i].cl_len = ski[i].struct_size = S_UNIX;

			/* crappy way, I know, I know ;) */
			ski[i].struct_size = strlen(ski[i].dst.path) +
				sizeof (ski[i].sa.unices.sun_family);
		}

		/* http://www.manualy.sk/sock-faq/unix-socket-faq-2.html#time_wait */
		if (setsockopt (ski[i].sd, SOL_SOCKET, SO_REUSEADDR,
				(char *)&sockopt, sizeof(sockopt)) == -1)
			perror(" (G-) setsockopts: REUSE");
		if (!makeitnonblocking (ski[i].sd))
			gprint("!socket [%d] is not nonblocking!\n", i);

		if (bind (ski[i].sd, ski[i].psa, ski[i].struct_size)) {
			perror (" (G!) bind()");
			gprint("-socket [%d] won't be used!\n", i);
			close(ski[i].sd);
			ski[i].sd = -1;
		}
	}
	return 0;
}

int make_listen(void)
{
	int i;
	/* MAKE'EM LISTEN */
	for (i=0; i<PORTS_NO; i++)
	{
		if (ski[i].sd == -1)
			continue;

		if (ski[i].proto != SOCK_STREAM)
			continue;

		if (listen (ski[i].sd, 5)) {
			perror (" (G!) listen()");
			gprint("-socket [%d] won't be used!\n", i);
			close(ski[i].sd);
			ski[i].sd = -1;
		}
	}
	return 0;
}

int printdata(int fd, struct sockaddr *psx, int fam, int ist)
{
	struct sockaddr_in  *gin_4 = (struct sockaddr_in  *)psx;
	struct sockaddr_in6 *gin_6 = (struct sockaddr_in6 *)psx;
	struct sockaddr_un  *gin_u = (struct sockaddr_un  *)psx;
	char locbuf[MBLOCK];

	if (fam == PF_INET) {
		inet_ntop(PF_INET, &(gin_4->sin_addr), locbuf, MBLOCK-1);
		gprint("+%s connection sd %d from %s:%d\n", ist?"TCP":"UDP",
				fd, locbuf, ntohs(gin_4->sin_port));
	} else if (fam == PF_INET6) {
		inet_ntop(PF_INET6, &(gin_6->sin6_addr), locbuf, MBLOCK-1);
		gprint("+%s Connection sd %d from %s:%d\n",ist?"TCP":"UDP",
				fd, locbuf, ntohs(gin_6->sin6_port));
	} else {
		gprint("+%s Connection sd %d from %s\n", ist?"TCP":"UDP",
				fd, gin_u->sun_path);
	}
	return 0;
}

int main(int argc, char **argv)
{
	int i, j, cd, ret;
	struct	timeval t;
	g_simlist iter, temp_iter;

#ifdef HAVE_LIBWRAP
	struct request_info	req;
#endif

	/* FIRST, WE START SIGINT SIGNAL HANDLER */
	signal (SIGINT, sigint_handler);
	prepare_and_bind_sockets();
	gprint("+Socket(s) prepared and binded\n");
	make_listen();
	gprint("+Socket(s) switched to listen mode\n");

	for (i=0; i<PORTS_NO; i++)
		if (ski[i].sd != -1 && ski[i].sd != 0)
			break;

	if (i == PORTS_NO) {
		gprint("!There is no ready socket!\n");
		exit(1);
	}

	signal (SIGCHLD, sigchld_handler);
	while (1) {
		/* ADDING TO rd AND wr SETS AND FINDING highsock */
		clearselect ();

		t.tv_sec = 1;
		t.tv_usec = 0;
		/* man 2 select
		 * "n  is the highest-numbered descriptor in any
		 *  of the three sets, plus 1."
		 */
		ret = select (highsock + 1, &rd, &wr, (fd_set *) 0, &t);
		if (ret == -1) {
			perror ("select()");
			shutdown_server();
			exit (-1);
		} else if (ret > 0) {
			for (i=0; i<PORTS_NO; i++)
			{
				cd = -1;
				if (ski[i].sd == -1)
					continue;
				/* IS SOMEONE CONNECTING TO US ? */
				if (FD_ISSET (ski[i].sd, &rd)) {
					ski[i].cl_len = ski[i].struct_size;
					if (ski[i].proto == SOCK_STREAM) {
						cd = accept (ski[i].sd, ski[i].psa, &ski[i].cl_len);
						if (cd < 0) {
							perror ("accept()");
							continue;
						}
#ifdef HAVE_LIBWRAP
						if (ski[i].family != PF_UNIX) {
							request_init(&req,
									RQ_FILE, cd,
									RQ_CLIENT_SIN, ski[i].psa,
									RQ_SERVER_SIN, &(ski[i].sa),
									RQ_DAEMON, ski[i].daemon,
								NULL);

							if(!hosts_access(&req))
							{
								gprint("!rejected by rules in hosts.allow,host.deny\n");
								close(cd);
								continue;
							}
						}
#endif
						makeitnonblocking (cd);
					}

					if (ski[i].proto == SOCK_STREAM) {
						for (j=0; j<MAXCLIENTS; j++)
							if (!ski[i].dzieci[j]) {
								printdata(cd, ski[i].psa, ski[i].family, 1);
								break;
							}
						if (j != MAXCLIENTS)
						{
							if (!(ski[i].dzieci[j] = fork())) {
								potomek(i, cd);
								exit (0);
							} else if (ski[i].dzieci[j] > 0) {
								close (cd);
								cd = -1;
							} else {
								perror(" (G!) fork()");
								gprint("-fork failed, rejecting client\n");
								if (ski[i].proto == SOCK_STREAM) 
									send (cd, "server is busy\n", 16, 0);
								close(cd);
								cd = -1;
							}
						}
					} else {
						cd = ski[i].sd;
						ski[i].klient=cd;
						ski[i].cl_len = sizeof(struct sockaddr_un);
						if(!odbieramy(i, ski[i].pclsa, &(ski[i].cl_len))) {
							iter = ski[i].kolejka_out;
							while (iter)
							{
								printdata(cd, ski[i].pclsa, ski[i].family, 0);
								if ((cd=sendto(cd, iter->string, 
										strlen(iter->string)+1, 0,
										ski[i].pclsa,
										ski[i].cl_len)) < 0) {
									perror(" (G!) sendto()");
								} else if (cd < strlen(iter->string)+1)
									gprint("-Less data sent then should be\n");
								else
									gprint("+Data send : %s", iter->string);
								temp_iter = iter->next;
								free(iter->string);
								free(iter);
								iter = temp_iter;
							}
							ski[i].kolejka_out = ski[i].kolejka_out_tail = NULL;
							cd = -1;
						}
					}
				}
				if (cd != -1) {
					gprint("-someone tried to connect, but we reject...\n");
					if (ski[i].proto == SOCK_STREAM)
						send (cd, "too many clients!\n", 19, 0);
					close (cd);
					cd = -1;
				}
			}
		}
	}
}

void sigchld_handler (int sig)
{
	int             pid;
	int		i, j;

	while ((pid = waitpid (-1, NULL, WNOHANG)) > 0);
	for (i=0; i<PORTS_NO; i++)
	{
		for(j=0; j<MAXCLIENTS; j++)
			if (ski[i].dzieci[j] == pid)
				ski[i].dzieci[j] = 0;
	}
}

void sigint_handler (int sig)
{
	shutdown_server();
	fprintf (stderr, " *** SERVER SHUT DOWN ***\n");
	exit (0);
}

void sigint_child_handler (int sig)
{
	gprint("+exiting child %d\n", getpid());
	close(highsock);
	exit (0);
}
/* THIS CLEAR ALL EMPTY SETS, ADD PROPER STUFF TO SETS,
 * AND FINDS highsock [THAT IS SOCKET WITH HIGHEST DESCRIPTOR]
 */
int clearselect (void)
{
	int             i;

	FD_ZERO (&rd);
	FD_ZERO (&wr);

	for (i=0; i<PORTS_NO; i++)
	{
		if (ski[i].sd == -1)
			continue;

		FD_SET (ski[i].sd, &rd);
		FD_SET (ski[i].sd, &wr);
		if (ski[i].sd > highsock)
			highsock = ski[i].sd;
	}
	return 0;
}

int potomek(int serv, int fd)
{
	struct timeval tv;
	g_simlist iter, temp_iter;
	int l, sl;

	signal(SIGINT, sigint_child_handler);
	ski[serv].klient = fd;
	highsock = fd;
	while (1)
	{
		FD_ZERO(&rd);
		FD_ZERO(&wr);
		FD_SET (fd, &rd);
		FD_SET (fd, &wr);

		tv.tv_sec = 0;
		tv.tv_usec = 1000;

		if (select(fd+1, &rd, &wr, NULL, &tv) > 0)
		{
			if (FD_ISSET(fd, &rd)) {
				if (odbieramy(serv, NULL, 0)) {
					close (fd);
					return 1;
				}
			}
			if (FD_ISSET(fd, &wr)) {
				iter = ski[serv].kolejka_out;
				while (iter)
				{
					l = strlen(iter->string);
					if ((sl = send(fd, iter->string, l+1, 0)) < 0) {
						perror(" (G!) send()");
					} else if (sl < l+1) {
						gprint("-Less data sent then should be\n");
					} else {
						gprint("+Data send : %s", iter->string);
					}
					temp_iter = iter->next;
					free(iter->string);
					free(iter);
					iter = temp_iter;
				}
				ski[serv].kolejka_out = ski[serv].kolejka_out_tail = NULL;
			}
		}
	}
	return 0;
}


/* THIS IS RESPONSIBLE FOR RECEIVING DATA FROM SOCKETS,
 * DATA BUFFERING AND CALLING PROPER FUNCTION, DEPENDING
 * ON x PARAMETER
 */
int odbieramy (int x, struct sockaddr *osin, int *osinlen)
{
	char            buforek[MBLOCK], tempchar, *p;
	int             r;

	if ((r = recvfrom (ski[x].klient, buforek, MBLOCK-1, 0, osin, osinlen)) <= 0) {
		/* operacja udana, pacjent zmar ;) */
		gprint("+Doctor, we've lost patient [%d]: %d\n", x, ski[x].klient);
		close (ski[x].klient);
		ski[x].klient = 0;
		ski[x].pbuf = 0;
		return 1;
	} else {
		buforek[r] = '\0';

		/* I'M DOING SUCH A BUFFERING STUFF, TO BE
		 * 'WINDOWS-COMPATIBILE', COUSE BY DEFAULT
		 * TELNET AND PUTTY UNDER WINDOWS SENDS
		 * BYTE BY BYTE, NOT WHOLE LINE
		 *
		 * I COULD ALSO ADD SOME 'BACKSPACE' STUFF
		 * HERE, BUT I'M NOT SURE IF THIS IS GOOD IDEA
		 */
		strncpy(ski[x].bufor+ski[x].pbuf,
				buforek, MBLOCK-ski[x].pbuf);
		ski[x].pbuf+=r;
		/* I'M SEARCHING IN ski* NOT IN buforek CAUSE i
		 * NEED POINTER TO SPECIFIC PLACE...
		 */
		while ((p=strchr(ski[x].bufor, '\n')) || (p=strchr(ski[x].bufor, '\r')))
		{
			p++;
			tempchar = *p;
			*p = '\0';
			gprint("+Received  : %s", ski[x].bufor);
			*p = tempchar;
			ski[x].handler(x);
			/* this strlen isn't probably good idea ;) */
			memmove(ski[x].bufor, p, strlen(p)+1);
			ski[x].pbuf = 0;
		}
	}
	return 0;
}

/*
 * HERE ARE THREE SIMPLE FUNCTIONS: dodawanie, mnozenie, leetspeech
 * WHICH SENDS BACK PROPER STUFF TO CLIENT
 */
int dodawanie (int a)
{
	char *buf=ski[a].bufor, *p=buf;
	char bufor_out[MBLOCK], *bufo=bufor_out;
	int x, y;

	if (!gatoi(&p, &x))
	{
		if (p)
			while (isspace((int)(*p))) p++;

		if (*p++ == '+') 
		{
			if (!gatoi(&p, &y)) {
				sprintf(bufo, "%d\n", x + y);
				if (add_to_queue(a, bufo))
					gprint("+Queued    : %s", bufo);
				else
					gprint("-Won't be send!\n    NOT queued: %s", bufo);
				return 0;
			}
		}
	}
	strcpy(bufo, "data format should be: a + b\n");
	if (add_to_queue(a, bufo))
		gprint("+Queued    : %s", bufo);
	else
		gprint("-Won't be send!\n    NOT queued: %s", bufo);
	return 1;
}

int mnozenie (int a)
{
	char *buf=ski[a].bufor, *p=buf;
	char bufor_out[MBLOCK], *bufo=bufor_out;
	int x, y;

	if (!gatoi(&p, &x))
	{
		if (p)
			while (isspace((int)(*p))) p++;

		if (*p++ == '*') 
		{
			if (!gatoi(&p, &y)) {
				sprintf(bufo, "%d\n", x * y);
				if (add_to_queue(a, bufo))
					gprint("+Queued    : %s", bufo);
				else
					gprint("-Won't be send!\nNOT queued: %s", bufo);
				return 0;
			}
		}
	}
	strcpy(bufo, "data format should be: a * b\n");
	if (add_to_queue(a, bufo))
		gprint("+Queued    : %s", bufo);
	else
		gprint("-Won't be send!\nNOT queued: %s", bufo);
	return 1;
}

int leetspeech(int a)
{
	char *buf=ski[a].bufor, *p, *q;
	char bufor_out[MBLOCK], *bufo=bufor_out, *o=bufo;
	int  c;
	for (p=buf, *o=*p; (*o=*p); ++p, ++o)
		if (isalpha((int)(*p)))
		{
			if (isupper((int)(*p)))
				c = tolower(*p);
			else
				c = *p;
			q=strchr(nor, c);
			if (q)
				*o=lit[q-nor];
		} 

	if (add_to_queue(a, bufo))
		gprint("+Queued    : %s", bufo);
	else
		gprint("-Won't be send!\nNOT queued: %s", bufo);
	return 0;
}

/* HMM GUESS IT'S HARD TO WORK-OUT WHAT THIS 
 * OBFUSCATED STUFF DOES ;)
 */
int shutdown_server(void)
{
	int i, j;
	for (i=0; i<PORTS_NO; i++)
	{
		if (ski[i].klient)
			close(ski[i].klient);
		for (j=0; j<MAXCLIENTS; j++)
			if (ski[i].dzieci[j]) {
				kill(SIGINT, ski[i].dzieci[j]);
				waitpid(ski[i].dzieci[j], NULL, 0);
			}
	}
	return closesockets ();
}

int closesockets(void)
{
	int i, j=0;
	for (i=0; i<PORTS_NO; i++) {
		if (ski[i].sd != 0 && ski[i].sd != -1)
			j |= close (ski[i].sd);
		if (ski[i].family == PF_UNIX)
			unlink (ski[i].dst.path);
	}

	return j;
}

/* SIMPLE WRAPPER AROUND strtol()
 */
int gatoi(char **buf, int *a)
{
	char *x;
	long t;
	t=strtol(*buf, &x, 10);
	if (x == *buf) return (1);
	*a = t;
	*buf = x;
	return (0);
}

/* COOL NONBLOCKING STUFF STOLEN FROM 
 * relayd.c (c) 1996,1997 Marcus Meissner <marcus@mud.de> */
int makeitnonblocking (int fd)
{
	int             isnonblock = 0;
	int             flags;

#if defined(FIONBIO)
	if (!isnonblock) {
		flags = 1;
		if (ioctl (fd, FIONBIO, &flags) == -1) {
			perror ("ioctl FIONBIO");
		} else
			isnonblock = 1;
	}
#elif defined(O_NDELAY)
	if (!isnonblock) {
		if ((flags = fcntl (fd, F_GETFL)) == -1) {
			perror ("fcntl F_GETFL");
		} else {
			flags |= O_NDELAY;
			if (fcntl (fd, F_SETFL, flags) == -1) {
				perror ("fcntl F_SETFL  O_NDELAY");
			} else
				isnonblock = 1;
		}
	}
#elif defined(O_NONBLOCK)
	if (!isnonblock) {
		if ((flags = fcntl (fd, F_GETFL)) == -1) {
			perror ("fcntl F_GETFL");
		} else {
			flags |= O_NONBLOCK;
			if (fcntl (fd, F_SETFL, flags) == -1) {
				perror ("fcntl F_SETFL  O_NONBLOCK");
			} else
				isnonblock = 1;
		}
	}
#endif
	return isnonblock;
}


g_simlist add_to_queue(int i, char *string)
{
	g_simlist nowy;
	nowy = (g_simlist)calloc(1, sizeof(struct g_simlist_));
	if (!nowy)
	{
		perror("Allocation error: calloc()");
		return NULL;
	}
	nowy->next = NULL;
	nowy->string = (char *)calloc(strlen(string)+1, 1);
	if (!(nowy->string))
	{
		free (nowy);
		perror("Allocation error: calloc()");
		return NULL;
	}
	memcpy(nowy->string, string, strlen(string)+1);
	if (!ski[i].kolejka_out)
		ski[i].kolejka_out = nowy;

	if (ski[i].kolejka_out_tail)
		ski[i].kolejka_out_tail->next = nowy;

	ski[i].kolejka_out_tail = nowy;
	return nowy;
}

int gprint(char *msg, ...)
{
	va_list ap;
	char buf[1024], *r=buf, *p = msg+1;

	va_start(ap, msg);
	while (*p)
	{
		if (*p == '%') {
			p++;
			if (*p == 's') {
				strcpy(r, va_arg(ap, char *));
				r += strlen(r);
			} else if (*p == 'd') {
				sprintf(r, "%d", va_arg(ap, int));
				r += strlen(r);
			} else {
				fprintf(stderr, " FATAL PROGRAMMER ERROR\n!");
				exit (1);
			}
		} else {
			*r = *p;
			r++;
		}
		p++;
	}
	*r = *p;
	va_end(ap);

	return fprintf(stderr, " (G%c) %s", *msg, buf);
}
